import asyncio
import inspect
import random
from functools import wraps
from typing import TYPE_CHECKING, Union, List, Dict

try:
    import aio_pika
    from aio_pika import ExchangeType, DeliveryMode
    from aio_pika.exceptions import AMQPConnectionError, ChannelClosed, QueueEmpty
except ImportError as e:
    raise ImportError(
        "Missing aio_pika dependencies. "
        "Please install: pip install aio_pika"
    ) from e

from tenacity import retry, wait_fixed, retry_if_exception_type

if TYPE_CHECKING:
    from ..crawler import Crawler

def auto_retry(func):
    @wraps(func)
    @retry(
        wait=wait_fixed(1),
        retry=retry_if_exception_type((AMQPConnectionError, ChannelClosed)),
        reraise=True
    )
    async def wrapper(self, *args, **kwargs):
        if self.stop_event.is_set():
            raise asyncio.CancelledError("Stop event set, abort RabbitMQ operation")
        try:
            return await func(self, *args, **kwargs)
        except (AMQPConnectionError, ChannelClosed):
            if self.stop_event.is_set():
                raise asyncio.CancelledError("Stop event set during reconnect")
            await self.connect()
            return await func(self, *args, **kwargs)
    return wrapper


class RabbitMQManager:
    def __init__(
        self,
        stop_event: asyncio.Event = None,
        rabbitmq_url: Union[str, List[str]] = None,
        exchange_name: str = "scrapy_cffi",
        exchange_type: ExchangeType = ExchangeType.DIRECT,
        prefetch_count: int = 10,
        persist: bool = False,
        loop: asyncio.AbstractEventLoop = None,
    ):
        self.stop_event = stop_event or asyncio.Event()
        self.exchange_name = exchange_name
        self.exchange_type = exchange_type
        self.prefetch_count = prefetch_count
        self.persist = persist
        self.loop = loop or asyncio.get_running_loop()
        self._lock = asyncio.Lock()

        if isinstance(rabbitmq_url, str):
            self.mq_mode = "single"
            self._mq_nodes = [rabbitmq_url]
        elif isinstance(rabbitmq_url, list):
            self.mq_mode = "cluster"
            if not rabbitmq_url:
                raise ValueError("Empty rabbitmq_url cluster node list")
            self._mq_nodes = rabbitmq_url
        else:
            raise ValueError("rabbitmq_url must be str or list of str")

        self._mq_url: str = None
        self._connection: aio_pika.RobustConnection = None
        self._channel: aio_pika.RobustChannel = None
        self._exchange: aio_pika.Exchange = None
        self._queues: Dict[str, aio_pika.Queue] = {}
        self._method_cache: Dict[str, callable] = {}

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            rabbitmq_url=crawler.settings.RABBITMQ_INFO.resolved_url,
            exchange_name=crawler.settings.RABBITMQ_INFO.EXCHANGE_NAME,
            exchange_type=crawler.settings.RABBITMQ_INFO.EXCHANGE_TYPE,
            prefetch_count=crawler.settings.RABBITMQ_INFO.PREFETCH_COUNT,
            persist=crawler.settings.SCHEDULER_PERSIST,
        )

    async def connect(self):
        self._mq_url = random.choice(self._mq_nodes)
        self._connection = await aio_pika.connect_robust(self._mq_url, loop=self.loop)
        self._channel = await self._connection.channel()
        await self._channel.set_qos(prefetch_count=self.prefetch_count)
        self._exchange = await self._channel.declare_exchange(
            self.exchange_name, type=self.exchange_type, durable=self.persist
        )

    def __getattribute__(self, name: str):
        if name.startswith("_") or name in ("_method_cache", "connect", "close"):
            return super().__getattribute__(name)

        attr = super().__getattribute__(name)
        if not callable(attr) or not inspect.iscoroutinefunction(attr):
            return attr

        method_cache = super().__getattribute__("_method_cache")
        if name not in method_cache:
            @wraps(attr)
            async def wrapper(*args, **kwargs):
                if self.stop_event.is_set():
                    raise asyncio.CancelledError(f"Stop event set, abort RabbitMQ operation: {name}")
                try:
                    return await attr(*args, **kwargs)
                except (AMQPConnectionError, ChannelClosed):
                    await self.connect()
                    return await attr(*args, **kwargs)
            method_cache[name] = wrapper
        return method_cache[name]

    @auto_retry
    async def declare_queue(self, queue_name: str, routing_key: str = None):
        if queue_name in self._queues:
            return self._queues[queue_name]
        queue = await self._channel.declare_queue(
            queue_name, durable=self.persist, auto_delete=not self.persist
        )
        routing_key = routing_key or queue_name
        await queue.bind(self._exchange, routing_key=routing_key)
        self._queues[queue_name] = queue
        return queue

    @auto_retry
    async def rpush(self, queue_name: str, message: bytes, routing_key: str = None):
        if self.stop_event.is_set():
            raise asyncio.CancelledError("Stop event set, abort RabbitMQ push")
        if not self._exchange:
            await self.connect()
        routing_key = routing_key or queue_name
        async with self._lock:
            await self._exchange.publish(
                aio_pika.Message(
                    body=message,
                    delivery_mode=DeliveryMode.PERSISTENT if self.persist else None
                ),
                routing_key=routing_key
            )
        return True

    @auto_retry
    async def dequeue_request(self, queue_name: str, timeout: int = 30) -> Union[bytes, None]:
        """
        Channel closed by RPC timeout.

        ⚠️ Warning: Do NOT set this timeout too low. This is a design flaw in aio_pika.
        If the timeout triggers, the channel may be closed, which can crash the entire program.
        """
        if self.stop_event.is_set():
            raise asyncio.CancelledError("Stop event set, abort RabbitMQ pop")
        if not self._exchange:
            await self.connect()
        queue = await self.declare_queue(queue_name)
        try:
            async with self._lock:
                message: aio_pika.IncomingMessage = await asyncio.wait_for(
                    queue.get(timeout=timeout),
                    timeout=timeout
                )
                if message:
                    async with message.process():
                        return message.body
                return None
        except asyncio.TimeoutError:
            return None
        except QueueEmpty:
            return None

    @auto_retry
    async def llen(self, queue_name: str) -> int:
        try:
            queue = await self._channel.declare_queue(
                queue_name, durable=self.persist, passive=True
            )
            return queue.declaration_result.message_count
        except aio_pika.exceptions.ChannelClosed:
            await self.connect()
            return 0
        except aio_pika.exceptions.ChannelInvalidStateError:
            return 0

    async def close(self):
        self.stop_event.set()
        try:
            if self._channel and not self._channel.is_closed:
                await self._channel.close()
        except ChannelClosed:
            pass
        try:
            if self._connection and not self._connection.is_closed:
                await self._connection.close()
        except AMQPConnectionError:
            pass



import asyncio
import functools
import json
from ..extensions import signals, Extension
from typing import TYPE_CHECKING, Callable, Awaitable, Any
if TYPE_CHECKING:
    from scrapy_cffi.hooks.signals import SignalsHooks
    from scrapy_cffi.extensions import SignalInfo

def locked(method: Callable[..., Awaitable[Any]]):
    @functools.wraps(method)
    async def wrapper(self, *args, **kwargs):
        async with self.lock:
            return await method(self, *args, **kwargs)
    return wrapper

class StatisticsExtension(Extension):
    count = {
        "engine_started": 0,
        "engine_stopped": 0,
        "scheduler_empty": 0,
        "task_error": 0,
        "spider_opened": 0,
        "spider_closed": 0,
        "spider_error": 0,
        "request_scheduled": 0,
        "request_dropped": 0,
        "request_reached_downloader": 0,
        "response_received": 0,
        "item_scraped": 0,
    }
    lock = asyncio.Lock()

    @classmethod
    def from_crawler(cls, hooks: "SignalsHooks"):
        extension_cls = cls(hooks)
        hooks.signals.connect(signals.engine_started, extension_cls.engine_started)
        hooks.signals.connect(signals.engine_stopped, extension_cls.engine_stopped)
        hooks.signals.connect(signals.scheduler_empty, extension_cls.scheduler_empty)
        hooks.signals.connect(signals.task_error, extension_cls.task_error)
        hooks.signals.connect(signals.spider_opened, extension_cls.spider_opened)
        hooks.signals.connect(signals.spider_closed, extension_cls.spider_closed)
        hooks.signals.connect(signals.spider_error, extension_cls.spider_error)
        hooks.signals.connect(signals.request_scheduled, extension_cls.request_scheduled)
        hooks.signals.connect(signals.request_dropped, extension_cls.request_dropped)
        hooks.signals.connect(signals.request_reached_downloader, extension_cls.request_reached_downloader)
        hooks.signals.connect(signals.response_received, extension_cls.response_received)
        hooks.signals.connect(signals.item_scraped, extension_cls.item_scraped)
        return extension_cls

    @locked
    async def engine_started(self, data: "SignalInfo"):
        self.count["engine_started"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def engine_stopped(self, data: "SignalInfo"):
        self.count["engine_stopped"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def scheduler_empty(self, data: "SignalInfo"):
        self.count["scheduler_empty"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def task_error(self, data: "SignalInfo"):
        self.count["task_error"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def spider_opened(self, data: "SignalInfo"):
        self.count["spider_opened"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def spider_closed(self, data: "SignalInfo"):
        self.count["spider_closed"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def spider_error(self, data: "SignalInfo"):
        self.count["spider_error"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def request_scheduled(self, data: "SignalInfo"):
        self.count["request_scheduled"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def request_dropped(self, data: "SignalInfo"):
        self.count["request_dropped"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def request_reached_downloader(self, data: "SignalInfo"):
        self.count["request_reached_downloader"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def response_received(self, data: "SignalInfo"):
        self.count["response_received"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))

    @locked
    async def item_scraped(self, data: "SignalInfo"):
        self.count["item_scraped"] += 1
        print(json.dumps(self.count, indent=4, ensure_ascii=False))


# utils/fixed.py
import struct
import binascii
from ..exceptions import DecoderException, EncoderException
from typing import Any, Tuple

# Generic functions for encoding/decoding structs based on the "struct" format
def encode_struct(fmt: str, value: Any) -> bytes:
    """Generic method for encoding arbitrary python "struct" values"""
    try:
        return struct.pack(fmt, value)
    except struct.error as exc:
        raise EncoderException(
            "Error encoding value %r with format string %s" % (value, fmt)
        )

def decode_struct(fmt: str, buf: bytes, pos: int) -> Tuple[Any, int]:
    """Generic method for decoding arbitrary python "struct" values"""
    new_pos = pos + struct.calcsize(fmt)
    try:
        return struct.unpack(fmt, buf[pos:new_pos])[0], new_pos
    except struct.error as exc:
        raise DecoderException(
            "Error deocding format string %s from bytes: %r"
            % (fmt, binascii.hexlify(buf[pos:new_pos]))
        )

def encode_fixed32(value: Any) -> bytes:
    """Encode a single 32 bit fixed-size value"""
    return encode_struct("<I", value)

def decode_fixed32(buf: bytes, pos: int) -> Tuple[Any, int]:
    """Decode a single 32 bit fixed-size value"""
    return decode_struct("<I", buf, pos)

def encode_sfixed32(value: Any) -> bytes:
    """Encode a single signed 32 bit fixed-size value"""
    return encode_struct("<i", value)

def decode_sfixed32(buf: bytes, pos: int) -> Tuple[Any, int]:
    """Decode a single signed 32 bit fixed-size value"""
    return decode_struct("<i", buf, pos)

def encode_float(value: Any) -> bytes:
    """Encode a single 32 bit floating point value"""
    return encode_struct("<f", value)

def decode_float(buf: bytes, pos: int) -> Tuple[Any, int]:
    """Decode a single 32 bit floating point value"""
    return decode_struct("<f", buf, pos)

def encode_fixed64(value: Any) -> bytes:
    """Encode a single 64 bit fixed-size value"""
    return encode_struct("<Q", value)

def decode_fixed64(buf: bytes, pos: int) -> Tuple[Any, int]:
    """Decode a single 64 bit fixed-size value"""
    return decode_struct("<Q", buf, pos)

def encode_sfixed64(value: Any) -> bytes:
    """Encode a single signed 64 bit fixed-size value"""
    return encode_struct("<q", value)

def decode_sfixed64(buf: bytes, pos: int) -> Tuple[Any, int]:
    """Decode a single signed 64 bit fixed-size value"""
    return decode_struct("<q", buf, pos)

def encode_double(value: Any) -> bytes:
    """Encode a single 64 bit floating point value"""
    return encode_struct("<d", value)

def decode_double(buf: bytes, pos: int) -> Tuple[Any, int]:
    """Decode a single 64 bit floating point value"""
    return decode_struct("<d", buf, pos)



from ..core import Request, HttpRequest, MediaRequest, WebSocketRequest, Response, HttpResponse, WebSocketResponse, CloseSignal
from ..models.api import WebSocketMsg

__all__ = [
    "Request",
    "HttpRequest",
    "MediaRequest",
    "WebSocketRequest",
    "Response",
    "HttpResponse",
    "WebSocketResponse",
    "CloseSignal",
    "WebSocketMsg",
]


import asyncio, re
from urllib.parse import urlparse
from curl_cffi import requests
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ..crawler import Crawler
    from ..settings import SettingsInfo
    from ..mq.kafka import KafkaManager

class RobotsTxtRules:
    def __init__(self, rules=None, fallback=False):
        self.rules = rules or []
        self.fallback = fallback

    def is_allowed(self, url: str) -> bool:
        parsed_url = urlparse(url)
        path = parsed_url.path or "/"

        matched_rule = None
        matched_length = -1

        for rule_type, rule_path in self.rules:
            if path.startswith(rule_path):
                if len(rule_path) > matched_length:
                    matched_length = len(rule_path)
                    matched_rule = rule_type
        return matched_rule != "disallow"

def parse_robots_txt(text: str, user_agent: str = "*") -> RobotsTxtRules:
    lines = text.splitlines()
    agent_blocks = {}
    current_agents = []

    for line in lines:
        line = line.strip()
        if not line or line.startswith("#"):
            continue

        match = re.match(r"(?i)^(user-agent|allow|disallow):\s*(.+)$", line)
        if not match:
            continue

        key, value = match.group(1).lower(), match.group(2).strip()

        if key == "user-agent":
            agent = value.lower()
            current_agents = [agent]
            if agent not in agent_blocks:
                agent_blocks[agent] = []
        elif key in ("allow", "disallow"):
            for agent in current_agents:
                agent_blocks.setdefault(agent, []).append((key, value))

    ua = user_agent.lower()
    matched_agents = [agent for agent in agent_blocks if agent in ua]
    if matched_agents:
        rules = []
        for agent in matched_agents:
            rules.extend(agent_blocks[agent])
        return RobotsTxtRules(rules)
    elif "*" in agent_blocks:
        return RobotsTxtRules(agent_blocks["*"])
    else:
        return RobotsTxtRules([])

class RobotsManager:
    def __init__(self, stop_event: asyncio.Event, settings: "SettingsInfo", kafkaManager: "KafkaManager"=None):
        self.stop_event = stop_event
        self.settings = settings
        from ..utils import init_logger
        self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        if kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)
        self._rules_cache = {}
        self._lock = asyncio.Lock()
        self._session = requests.AsyncSession()

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(stop_event=crawler.stop_event, settings=crawler.settings, kafkaManager=crawler.kafkaManager)

    async def load_rules_for_hosts(self, robot_urls):
        tasks = []
        async with self._lock:
            for url in robot_urls:
                domain = urlparse(url).netloc
                if domain not in self._rules_cache:
                    tasks.append(self._load_single(url, domain))
        await asyncio.gather(*tasks)
        await self._session.close()

    async def _load_single(self, url, domain):
        rules = RobotsTxtRules([], fallback=True)
        try:
            resp = await self._session.get(url, headers={
                "sec-ch-prefers-color-scheme": "light",
                "sec-ch-ua-mobile": "?0",
                "sec-fetch-dest": "document",
                "sec-fetch-mode": "navigate",
                "sec-fetch-site": "none",
                "sec-fetch-user": "?1",
                "user-agent": self.settings.USER_AGENT
            }, timeout=10)
            if resp.status_code == 200:
                text = resp.text
                rules = parse_robots_txt(text, self.settings.USER_AGENT)
                if not rules.rules:
                    self.logger.info(f"robots.txt from {url} has no rules, allowing all.")
                    rules = RobotsTxtRules([], fallback=True)
            elif resp.status_code == 404:
                self.logger.debug(f"robots.txt not found on {url}, allowing all.")
            else:
                self.logger.warning(f"robots.txt load failed from {url}, status: {resp.status_code}")
        except Exception as e:
            self.logger.warning(f"Error fetching robots.txt from {url}, disallowing all. -> {e}")
            rules = RobotsTxtRules([("disallow", "/")], fallback=False)

        async with self._lock:
            self._rules_cache[domain] = rules

    def is_allowed(self, url: str) -> bool:
        domain = urlparse(url).netloc
        rules: RobotsTxtRules = self._rules_cache.get(domain)
        return rules.is_allowed(url) if rules else True



import asyncio, time
from ..downloader.internet import Request
from typing import TYPE_CHECKING, List, Dict
# from ...utils import run_with_timeout
from ...extensions import signals, SignalInfo
from ..sessions import SessionManager
if TYPE_CHECKING:
    from ...crawler import Crawler
    from ...settings import SettingsInfo
    from ...spiders import Spider
    from ...extensions import SignalManager

class BaseScheduler:
    def __init__(
        self, 
        spiders_name: List=None,
        stop_event: asyncio.Event=None, 
        settings: "SettingsInfo"=None, 
        sessions: "SessionManager"=None, 
        sessions_lock: asyncio.Lock=None, 
        signalManager: "SignalManager"=None, 
        **kwargs
    ):
        self.spiders_name = spiders_name
        self.stop_event = stop_event
        self.settings = settings
        self.sessions = sessions
        self.sessions_lock = sessions_lock
        self.signalManager = signalManager
        self.kwargs = kwargs
        self.is_distributed = False

    @classmethod
    def from_crawler(cls, crawler: "Crawler", spiders_name: List):
        return cls(
            spiders_name=spiders_name, 
            stop_event=crawler.stop_event,
            settings=crawler.settings,
            sessions=crawler.sessions,
            sessions_lock=crawler.sessions_lock,
            signalManager=crawler.signalManager
        )
    
    def get_queue_key(self, spider: "Spider") -> str:
        return self.settings.QUEUE_NAME if self.settings.QUEUE_NAME else f"{spider.name}_req"
    
    async def put(self, request: Request, spider: "Spider", **kwargs):
        raise NotImplementedError

    async def get(self, spider: "Spider"=None, **kwargs):
        raise NotImplementedError

class Scheduler(BaseScheduler):
    def __init__(
        self, 
        spiders_name: List=None,
        stop_event: asyncio.Event=None, 
        settings: "SettingsInfo"=None, 
        sessions: "SessionManager"=None, 
        sessions_lock: asyncio.Lock=None, 
        signalManager: "SignalManager"=None, 
        **kwargs
    ):
        super().__init__(
            spiders_name=spiders_name, 
            stop_event=stop_event, 
            settings=settings, 
            sessions=sessions, 
            sessions_lock=sessions_lock, 
            signalManager=signalManager, 
            **kwargs
        )
        if self.settings.DUPEFILTER:
            from ...utils import load_object
            dupefilter_cls = load_object(path=self.settings.DUPEFILTER)
            self.dupefilter = dupefilter_cls(settings=self.settings, **kwargs)
        else:
            from ...dupefilter.base import MemoryDupeFilter
            self.dupefilter = MemoryDupeFilter(settings=self.settings, **kwargs)
        self._queue_map: Dict[str, asyncio.Queue] = {}
        if self.settings.QUEUE_NAME:
            self._queue_map[self.settings.QUEUE_NAME] = asyncio.Queue()
        else:
            for spider_name in self.spiders_name:
                self._queue_map[f"{spider_name}_req"] = asyncio.Queue()

    async def put(self, request: Request, spider: "Spider", **kwargs):
        # Requests with dont_filter=True or WebSocket requests signaling connection end should not be deduplicated
        if request.dont_filter:
            await self._queue_map[self.get_queue_key(spider=spider)].put(request)
            self.signalManager.send(signal=signals.request_scheduled, data=SignalInfo(signal_time=time.time(), request=request))
            return True
        else:
            async with self.dupefilter.lock:
                is_seen = await self.dupefilter.request_seen(request=request)
                if not is_seen:
                    await self._queue_map[self.get_queue_key(spider=spider)].put(request)
                    self.signalManager.send(signal=signals.request_scheduled, data=SignalInfo(signal_time=time.time(), request=request))
                    return True
                else:
                    async with self.sessions_lock:
                        self.sessions.release(session_id=request.session_id)
                    self.signalManager.send(signal=signals.request_dropped, data=SignalInfo(signal_time=time.time(), request=request, reason=f"filter: {request.url}"))
                    return False

    async def put_is_req(self, request: "Request", spider: "Spider", **kwargs):
        return await self.dupefilter.mark_sent(request=request, spider=spider, **kwargs)

    async def get(self, spider: "Spider"=None, **kwargs):
        return await self._queue_map[self.get_queue_key(spider=spider)].get()

    def empty(self, spider: "Spider", **kwargs) -> bool:
        return self._queue_map[self.get_queue_key(spider=spider)].empty()


import asyncio
import inspect
from functools import wraps
from typing import Optional, Any, Callable, Coroutine, TYPE_CHECKING
try:
    from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorCollection
except ImportError as e:
    raise ImportError(
        "Missing motor dependencies. "
        "Please install: pip install motor>=3.7.1"
    ) from e
if TYPE_CHECKING:
    from ..crawler import Crawler

RETRYABLE_EXCEPTIONS = (ConnectionError, TimeoutError)

def auto_retry(func: Callable[..., Coroutine[Any, Any, Any]]):
    @wraps(func)
    async def wrapper(self, *args, **kwargs):
        if self.stop_event.is_set():
            raise asyncio.CancelledError("Stop event set, abort MongoDB operation")
        try:
            return await func(self, *args, **kwargs)
        except RETRYABLE_EXCEPTIONS:
            if self.stop_event.is_set():
                raise asyncio.CancelledError("Stop event set during reconnect")
            await self._reconnect()
            return await func(self, *args, **kwargs)
    return wrapper

class MongoCollectionWrapper:
    def __init__(self, collection: AsyncIOMotorCollection, manager: "MongoDBManager"):
        self._collection = collection
        self._manager = manager

    def __getattr__(self, name: str):
        attr = getattr(self._collection, name)

        if not callable(attr) or not inspect.iscoroutinefunction(attr):
            return attr

        @wraps(attr)
        async def wrapper(*args, **kwargs):
            if self._manager.stop_event.is_set():
                raise asyncio.CancelledError("Stop event set, abort MongoDB operation")
            try:
                return await attr(*args, **kwargs)
            except RETRYABLE_EXCEPTIONS:
                if self._manager.stop_event.is_set():
                    raise asyncio.CancelledError("Stop event set during reconnect")
                await self._manager._reconnect()
                new_attr = getattr(self._manager.db.get_collection(self._collection.name), name)
                return await new_attr(*args, **kwargs)
        return wrapper

class MongoDBManager:
    def __init__(self, stop_event: asyncio.Event, mongo_uri: str, db_name: Optional[str] = None):
        if not db_name:
            raise ValueError("MongoDBManager requires a valid db_name.")
        self.stop_event = stop_event
        self.mongo_uri = mongo_uri
        self.db_name = db_name
        self.client: Optional[AsyncIOMotorClient] = None
        self.db = None

    @classmethod
    def from_crawler(cls, crawler: "Crawler") -> "MongoDBManager":
        return cls(
            stop_event=crawler.stop_event,
            mongo_uri=crawler.settings.MONBODB_INFO.resolved_url,
            db_name=crawler.settings.MONBODB_INFO.DB,
        )

    async def _reconnect(self):
        if self.client:
            self.client.close()
        self.client = AsyncIOMotorClient(self.mongo_uri)
        self.db = self.client[self.db_name]

    async def init(self):
        await self._reconnect()

    async def close(self):
        if self.client:
            self.client.close()

    def collection(self, name: str) -> AsyncIOMotorCollection:
        return MongoCollectionWrapper(self.db.get_collection(name), self)

    @auto_retry
    async def list_collections(self):
        return await self.db.list_collection_names()
    
    @auto_retry
    async def drop_database(self, db_name: Optional[str] = None):
        await self.client.drop_database(db_name or self.db_name)



from scrapy_cffi.extensions import signals, Extension
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from scrapy_cffi.hooks.signals import SignalsHooks
    from scrapy_cffi.extensions import SignalInfo

class CustomExtension(Extension):
    @classmethod
    def from_crawler(cls, hooks: "SignalsHooks", **kwargs):
        extension_cls = cls(hooks, **kwargs)
        hooks.signals.connect(signals.engine_started, extension_cls.engine_started)
        hooks.signals.connect(signals.engine_stopped, extension_cls.engine_stopped)
        hooks.signals.connect(signals.scheduler_empty, extension_cls.scheduler_empty)
        hooks.signals.connect(signals.task_error, extension_cls.task_error)
        hooks.signals.connect(signals.spider_opened, extension_cls.spider_opened)
        hooks.signals.connect(signals.spider_closed, extension_cls.spider_closed)
        hooks.signals.connect(signals.spider_error, extension_cls.spider_error)
        hooks.signals.connect(signals.request_scheduled, extension_cls.request_scheduled)
        hooks.signals.connect(signals.request_dropped, extension_cls.request_dropped)
        hooks.signals.connect(signals.request_reached_downloader, extension_cls.request_reached_downloader)
        hooks.signals.connect(signals.response_received, extension_cls.response_received)
        hooks.signals.connect(signals.item_scraped, extension_cls.item_scraped)
        return extension_cls

    async def engine_started(self, data: "SignalInfo"):
        print(f'CustomExtension -> engine started: {data.signal_time}')

    async def engine_stopped(self, data: "SignalInfo"):
        print(f'CustomExtension -> engine stopped: {data.signal_time}')

    async def scheduler_empty(self, data: "SignalInfo"):
        print(f'CustomExtension -> scheduler empty: {data.signal_time}')

    async def task_error(self, data: "SignalInfo"):
        print(f'CustomExtension -> task error: {data.reason}, signal_time：{data.signal_time}')

    async def spider_opened(self, data: "SignalInfo"):
        print(f'CustomExtension -> spider opened: {data.spider.name}, start_time: {data.signal_time}')

    async def spider_closed(self, data: "SignalInfo"):
        print(f'CustomExtension -> spider closed: {data.spider.name}, end_time: {data.signal_time}')

    async def spider_error(self, data: "SignalInfo"):
        print(f'CustomExtension -> spider error: {data.spider.name}, exception: {data.exception}, signal_time：{data.signal_time}')

    async def request_scheduled(self, data: "SignalInfo"):
        print(f'CustomExtension -> request scheduled: {data.request.url}, signal_time：{data.signal_time}')

    async def request_dropped(self, data: "SignalInfo"):
        print(f'CustomExtension -> request dropped: {data.request.url}, reason: {data.reason}, signal_time：{data.signal_time}')

    async def request_reached_downloader(self, data: "SignalInfo"):
        print(f'CustomExtension -> request reached downloader: {data.request.url}, signal_time：{data.signal_time}')

    async def response_received(self, data: "SignalInfo"):
        print(f'CustomExtension -> response received: {data.response}, request: {data.request.url}, signal_time：{data.signal_time}')

    async def item_scraped(self, data: "SignalInfo"):
        print(f'CustomExtension -> item scraped: {data.item}, spider_name: {data.spider.name}, signal_time：{data.signal_time}')


import os, sys
from ctypes import CDLL, c_void_p, c_char_p, c_bool, c_size_t, POINTER
from pathlib import Path
from typing import Union

dir = Path(__file__).parent
if os.name.lower() == "nt":
    # Windows
    lib = CDLL(str(dir / "build" / "libbloom.dll"))
elif sys.platform == "darwin":
    # macOS
    lib = CDLL(str(dir / "build" / "libbloom.dylib"))
else:
    # Linux / Unix
    lib = CDLL(str(dir / "build" / "libbloom.so"))

lib.bloom_init.restype = c_void_p
lib.bloom_init.argtypes = [c_size_t, c_size_t, c_size_t]

lib.bloom_add.argtypes = [c_void_p, c_char_p, c_size_t]
lib.bloom_add.restype = None

lib.bloom_exists.argtypes = [c_void_p, c_char_p, c_size_t]
lib.bloom_exists.restype = c_bool

lib.bloom_free.argtypes = [c_void_p]

lib.bloom_clear.argtypes = [c_void_p]

lib.bloom_estimated_false_positive_rate.argtypes = [c_void_p]
lib.bloom_estimated_false_positive_rate.restype = float

lib.bloom_get_hash_count.restype = c_size_t
lib.bloom_get_hash_count.argtypes = [c_void_p]

lib.bloom_get_indices.argtypes = [c_void_p, c_char_p, c_size_t, POINTER(c_size_t)]
lib.bloom_get_indices.restype = c_size_t

# Python code
def bloom_init(size: int, expected: int, k: int = 0) -> c_void_p:
    return lib.bloom_init(size, expected, k)

def bloom_add(filter_ptr: c_void_p, data: Union[str, bytes]):
    if isinstance(data, str):
        data = data.encode()
    lib.bloom_add(filter_ptr, data, len(data))

def bloom_exists(filter_ptr: c_void_p, data: Union[str, bytes]) -> bool:
    if isinstance(data, str):
        data = data.encode()
    return lib.bloom_exists(filter_ptr, data, len(data))

def bloom_free(filter_ptr: c_void_p):
    lib.bloom_free(filter_ptr)

def bloom_clear(filter_ptr: c_void_p):
    lib.bloom_clear(filter_ptr)

def bloom_estimated_false_positive_rate(filter_ptr: c_void_p) -> float:
    return lib.bloom_estimated_false_positive_rate(filter_ptr)

def bloom_get_hash_count(filter_ptr: c_void_p) -> int:
    return lib.bloom_get_hash_count(filter_ptr)

def bloom_get_indices(filter_ptr: c_void_p, data: Union[str, bytes]) -> list[int]:
    if isinstance(data, str):
        data = data.encode()
    hash_count = bloom_get_hash_count(filter_ptr)
    indices_array = (c_size_t * hash_count)()
    count = lib.bloom_get_indices(filter_ptr, data, len(data), indices_array)
    return [indices_array[i] for i in range(count)]

class BloomFilter:
    def __init__(self, size: int, expected: int, hash_count: int = 0):
        self._ptr = bloom_init(size, expected, hash_count)

    def add(self, data: Union[str, bytes]):
        bloom_add(self._ptr, data)

    def exists(self, data: Union[str, bytes]) -> bool:
        return bloom_exists(self._ptr, data)
    
    def free(self):
        bloom_free(self._ptr)

    def clear(self):
        bloom_clear(self._ptr)

    def estimated_false_positive_rate(self) -> float:
        return bloom_estimated_false_positive_rate(self._ptr)

    def get_hash_count(self) -> int:
        return bloom_get_hash_count(self._ptr)

    def get_indices(self, data: Union[str, bytes]) -> list[int]:
        return bloom_get_indices(self._ptr, data)
    
    def __enter__(self):
        return self
    def __exit__(self, exc_type, exc_val, exc_tb):
        self.free()

__all__ = [
    "BloomFilter"
]

if __name__ == "__main__":
    bf = BloomFilter(size=1000, expected=100, hash_count=0)
    print(bf.get_hash_count())
    bf.add("asgshyjhyt")
    bf.add("asgyiylo98")
    bf.add("asgyiylo96687")
    print(bf.get_indices("daafasg"))


from functools import cached_property
from curl_cffi.requests import Response as CffiResponse
from ..selector import Selector
from ....utils import ProtobufFactory
from typing import Tuple, Dict, List, Union

class Response(object):
    def __init__(self,
        session_id="",
        raw_response: CffiResponse=None,
        meta={},
        dont_filter=None,
        callback=None,
        errback=None,
        desc_text="",
        request=None,
        **kwargs
    ) -> None:
        self.session_id = session_id
        self.raw_response = raw_response
        self.meta = meta
        self.dont_filter = dont_filter
        self.callback = callback
        self.errback = errback
        self.desc_text = desc_text
        self.request = request
        self.kwargs = kwargs

class HttpResponse(Response):
    def __init__(self, 
        session_id="",
        raw_response: CffiResponse=None,
        meta={},
        dont_filter=None,
        callback=None,
        errback=None,
        desc_text="",
        request=None,
        **kwargs
    ):
        super().__init__(
            session_id=session_id,
            raw_response=raw_response,
            meta=meta,
            dont_filter=dont_filter,
            callback=callback,
            errback=errback,
            desc_text=desc_text,
            request=request,
            **kwargs
        )
        self.status_code = self.raw_response.status_code     
        self.content = self.raw_response.content
        self.text = self.raw_response.text

    def get_selector_type(self):
        ctype = self.raw_response.headers.get("Content-Type", "").lower()
        if "xml" in ctype:
            return "xml"
        elif "html" in ctype:
            return "html"
        return "other"

    @cached_property
    def selector(self):
        return Selector(response=self.raw_response, type=self.get_selector_type())

    def xpath(self, query):
        return self.selector.xpath(query)
    
    def css(self, query):
        return self.selector.css(query)

    def re(self, pattern):
        return self.selector.re(pattern)
    
    def json(self):
        return self.selector.json()
    
    def extract_json(self, key: str="", re_rule: str="") -> Union[List[Union[Dict, str]], Dict, str]:
        return self.selector.extract_json(key, re_rule=re_rule)

    def extract_json_strong(self, key: str="", strict_level=2, re_rule="") -> Union[List[Union[Dict, str]], Dict, str]:
        return self.selector.extract_json_strong(key, strict_level=strict_level, re_rule=re_rule)
    
    def protobuf_decode(self) -> Tuple[Dict, Dict]:
        return self.selector.protobuf_decode()
    
    def grpc_decode(self) -> Union[Tuple[Dict, Dict], List[Tuple[Dict, Dict]]]:
        return self.selector.grpc_decode()
    
class WebSocketResponse(Response):
    def __init__(self, 
        session_id="",
        websocket_id="",
        msg=b'',
        meta={},
        callback=None,
        errback=None,
        desc_text="",
        request=None,
        **kwargs
    ):
        super().__init__(session_id=session_id, meta=meta, callback=callback, errback=errback, desc_text=desc_text, request=request, **kwargs)
        self.websocket_id = websocket_id
        self.msg = msg

    def protobuf_decode(self) -> Tuple[Dict, Dict]:
        return ProtobufFactory.protobuf_decode(self.msg)
    
    def grpc_decode(self) -> Union[Tuple[Dict, Dict], List[Tuple[Dict, Dict]]]:
        return ProtobufFactory.grpc_decode(self.msg)


from .chains import ChainManager, InterruptibleChainManager, ChainResult, ChainNextEnum
from .base import DownloadInterceptor, SpiderInterceptor

__all__ = [
    "ChainManager",
    "InterruptibleChainManager",
    "ChainResult",
    "ChainNextEnum",
    "DownloadInterceptor",
    "SpiderInterceptor",
]


import asyncio, random
from urllib.parse import urlparse
from typing import Union, AsyncGenerator
from collections.abc import AsyncIterable, Iterable
from .base import _InnerSpiderInterceptor
from ..spiders import Spider
from ..hooks import interceptors_hooks
from ..core.downloader.internet import *
from ..core.sessions import CloseSignal
from ..exceptions import SessionEndError, BlockRequestError, FilterDomainRequestError
from typing import TYPE_CHECKING, Dict
if TYPE_CHECKING:
    from ..crawler import Crawler
    from ..settings import SettingsInfo
    from ..utils import RobotsManager
    from ..item import Item
    from ..hooks.interceptors import InterceptorsHooks
    from ..core.sessions import SessionWrapper, WebSocketEntry
    from ..mq.kafka import KafkaManager

class UpdateRequestSpiderInterceptor(_InnerSpiderInterceptor):
    ResultType = Union[
        AsyncGenerator,
        AsyncIterable,
        Iterable,
        Request,
        Dict,
        None,
    ]

    def __init__(
        self, 
        stop_event: asyncio.Event=None, 
        settings: "SettingsInfo"=None, 
        hooks: "InterceptorsHooks"=None, 
        sessions_lock=None,
        kafkaManager: "KafkaManager"=None,
        **kwargs
    ):
        super().__init__(stop_event=stop_event, settings=settings, hooks=hooks, sessions_lock=sessions_lock, kafkaManager=kafkaManager, **kwargs)
        from ..utils import init_logger
        self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        if self.kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=self.kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)

        self.default_ua = self.settings.USER_AGENT
        self.default_headers = self.settings.DEFAULT_HEADERS
        self.timeout = self.settings.TIMEOUT
        self.dont_filter = self.settings.DONT_FILTER
        self.proxies = self.settings.PROXIES
        self.proxies_list = [{"http": proxy_url, "https": proxy_url} for proxy_url in self.settings.PROXIES_LIST]

    def pre_check(self, request: Request) -> Request:
        if not request.headers:
            request.headers = self.default_headers
        if self.default_ua:
            ua = request.find_header_key(key="user-agent")
            if not ua:
                request.headers["user-agent"] = self.default_ua

        if not request.timeout:
            request.timeout = self.timeout

        if request.no_proxy:
            request.proxies = None
        elif not request.proxies:
            if self.proxies:
                request.proxies = self.proxies
            elif self.proxies_list:
                request.proxies = random.choice(self.proxies_list)

        if request.dont_filter is None:
            request.dont_filter = True if self.dont_filter else False
        return request

    async def process_spider_output(self, response: Response, result: ResultType, spider: Spider):
        """
        Default Spider Middleware

        This middleware is the first in the spider middleware chain and serves as a foundational processor for all requests.

        Key responsibilities:
        - Complete and validate all requests before further processing.
        - Manage concurrency by locking sessions to prevent race conditions.
        - Ensure WebSocket requests correctly associate with existing WebSocket connections,
        handling session cookie updates and URL assignments.
        - Detect and handle closed WebSocket connections gracefully by returning errors and logging warnings.

        This middleware ensures that subsequent middleware and the engine receive well-formed,
        locked, and valid requests, especially important for managing WebSocket sessions.

        It plays a critical role in stabilizing the request lifecycle and maintaining session consistency.
        """
        if isinstance(result, (Request, CloseSignal)):
            async with self.sessions_lock:
                self.hooks.session.acquire(session_id=result.session_id)

                if isinstance(result, CloseSignal):
                    wrapper: "SessionWrapper" = self.hooks.session.get_or_create_session(session_id=result.session_id)
                    if result.websocket_end_for_key:
                        webSocket_entry: "WebSocketEntry" = wrapper.websocket_pool.get_from_key(result.websocket_end_for_key)
                        if webSocket_entry:
                            result.websocket_end_for_url = webSocket_entry.url
                            wrapper.websocket_pool.acquire_from_url(url=result.websocket_end_for_url)
                        
                    if (not result.websocket_end_for_url) and (not result.session_end):
                        self.logger.error(
                            f"Received CloseSignal with an unknown end tag — it will be ignored. "
                            f"(websocket_end_for_url: {result.websocket_end_for_url}, session_end: {result.session_end})"
                        )
                        return
                else:
                    result = self.pre_check(result)
                    wrapper: "SessionWrapper" = self.hooks.session.get_or_create_session(session_id=result.session_id, cookies=result.cookies)
                    if result.cookies:
                        wrapper.update_session_cookies(cookies_dict=result.cookies)

                    # For non-initial WebSocket requests:
                    # The websocket_pool should already contain the key, 
                    # so ensure the request.url is set accordingly and acquire a lock on it.
                    # For initial WebSocket requests, only the session lock is applied.
                    if isinstance(result, WebSocketRequest) and ((not result.url) and result.websocket_id):
                        webSocket_entry: "WebSocketEntry" = wrapper.websocket_pool.get_from_key(result.websocket_id)
                        if webSocket_entry:
                            result.url = webSocket_entry.url
                            wrapper.websocket_pool.acquire_from_url(url=result.url)
                        else:
                            error_text = f'WebSocket connection {result.websocket_id} has been closed, but a new WebSocketRequest was received'
                            self.logger.warning(f'{error_text}: {result.send_message}')
                            return SessionEndError(exception=ValueError(error_text), request=result)
        return result
    
class RobotSpiderInterceptor(_InnerSpiderInterceptor):
    def __init__(
        self,  
        stop_event: asyncio.Event=None, 
        settings: "SettingsInfo"=None, 
        hooks: "InterceptorsHooks"=None, 
        sessions_lock=None, 
        robot: "RobotsManager"=None,
        kafkaManager: "KafkaManager"=None,
        **kwargs
    ):
        super().__init__(stop_event=stop_event, settings=settings, hooks=hooks, sessions_lock=sessions_lock, kafkaManager=kafkaManager, **kwargs)
        self.robot = robot
        from ..utils import init_logger
        self.logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        if self.kafkaManager:
            from ..utils import KafkaLoggingHandler
            kafka_handler = KafkaLoggingHandler(kafka=self.kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)

    @classmethod
    def from_crawler(cls, crawler: "Crawler"):
        return cls(
            stop_event=crawler.stop_event,
            settings=crawler.settings,
            hooks=interceptors_hooks(crawler),
            sessions_lock=crawler.sessions_lock,
            robot=crawler.robot,
            kafkaManager=crawler.kafkaManager
        )

    def is_allow(self, url, allow_domains):
        parsed = urlparse(url)
        domain = parsed.netloc.lower()
        if not domain:
            return False
        domain = domain.lower()
        for allowed in allow_domains:
            allowed = allowed.lower()
            if domain == allowed or domain.endswith('.' + allowed):
                return True
        return False

    async def process_spider_output(self, response: Response, result: Union[Request, "Item", dict, None], spider: Spider):
        """
        Default second spider middleware.

        Responsibilities:
        - Filter out requests whose URLs are not within the spider's allowed domains.
        - Release the session lock for filtered requests.
        - Enforce robots.txt rules if enabled in settings.
        - Log warnings or debug messages when requests are filtered or blocked.
        - Return specific exceptions to indicate filtered or blocked requests.

        Parameters:
        - response: The HTTP response associated with the spider output.
        - result: The spider output item, typically a Request, Item, dict, or None.
        - spider: The spider instance.

        Returns:
        - The original result if allowed.
        - An error wrapper object if the request is filtered or blocked.
        """
        if isinstance(result, Request) and result.url:
            is_start_url = result.meta.pop("is_start_url", None)
            if not is_start_url:
                is_allow_res = self.is_allow(url=result.url, allow_domains=spider.allowed_domains)
                if not is_allow_res:
                    async with self.sessions_lock:
                        self.hooks.session.release(session_id=result.session_id)
                    log_text = f"Request URL {result.url} is outside allowed_domains and has been filtered"
                    self.logger.warning(log_text)
                    return FilterDomainRequestError(exception=ValueError(log_text), request=result)
                
                if self.settings.ROBOTSTXT_OBEY and (not self.robot.is_allowed(result.url)):
                    async with self.sessions_lock:
                        self.hooks.session.release(session_id=result.session_id)
                    log_text = f"[robots.txt] Blocked: {result.url}"
                    self.logger.debug(log_text)
                    return BlockRequestError(exception=ValueError(log_text), request=result)
        return result


from .models import BaseValidatedModel, StrictValidatedModel
from .models.api import ComponentInfo, RedisInfo, MysqlInfo, MongodbInfo, RabbitMQInfo, KafkaInfo, CPYExtensionsConfig
from pydantic import field_validator, model_validator, ValidationInfo, PrivateAttr, Field
from typing import Optional, List, Dict, Union, Any, ClassVar, Literal

class BloomInfo(StrictValidatedModel):
    MODE: Optional[bool] = False
    SIZE: Optional[int] = 100000000
    EXPECTED: Optional[int] = 100000000
    HASH_COUNT: Optional[int] = 0

class LogInfo(StrictValidatedModel):
    _encoding_fields: ClassVar[List[str]] = ["LOG_ENCODING"]

    LOG_ENABLED: Optional[bool] = True
    LOG_WITH_STREAM: Optional[bool] = True
    LOG_LEVEL: Optional[str] = "DEBUG"
    LOG_FORMAT: Optional[str] = "%(asctime)s [%(name)s] %(levelname)s: %(message)s"
    LOG_DATEFORMAT: Optional[str] = "%Y-%m-%d %H:%M:%S"
    LOG_FILE: Optional[str] = ""
    LOG_ENCODING: Optional[str] = "utf-8"
    LOG_SHORT_NAMES: Optional[bool] = False
    LOG_FORMATTER: Optional[str] = ""

    @field_validator('LOG_LEVEL')
    @classmethod
    def normalize_mute_type(cls, log_level: str):
        if log_level:
            valid_levels = {"DEBUG", "INFO", "WARNING", "ERROR", "CRITICAL"}
            if log_level.upper() not in valid_levels:
                raise ValueError(f'LOG_LEVEL must be one of {valid_levels}')
            return log_level.upper()
        return log_level

class SettingsInfo(BaseValidatedModel):
    # _encoding_fields: ClassVar[List[str]] = ["FEED_EXPORT_ENCODING"]

    MAX_GLOBAL_CONCURRENT_TASKS: Optional[Union[int, None]] = 300 # asyncio.BoundedSemaphore()
    QUEUE_NAME: Optional[Union[str]] = "" # If set, this queue will be shared in run_all_spiders mode. Be aware of potential request race conditions when using the same scheduler.
    ROBOTSTXT_OBEY: Optional[bool] = True # Whether to respect robots.txt rules

    USER_AGENT: Optional[str] = "scrapy_cffiBot"
    DEFAULT_HEADERS: Optional[Dict] = Field(default_factory=dict)
    DEFAULT_COOKIES: Optional[Dict] = Field(default_factory=dict)
    MAX_CONCURRENT_REQ: Optional[int] = None # asyncio.Semaphore()
    USE_STRICT_SEMAPHORE: Optional[bool] = False # asyncio.BoundedSemaphore()
    TIMEOUT: Optional[int] = 30 # Request timeout in seconds
    MAX_REQ_TIMES: Optional[int] = 2 # Maximum number of retry attempts for a failed request
    DELAY_REQ_TIME: Optional[int] = 3 # Delay in seconds before retrying a failed request
    
    PROXY_URL: Optional[str] = None
    PROXIES: Optional[Dict] = None
    PROXIES_LIST: Optional[List[str]] = Field(default_factory=list)
    
    SPIDERS_PATH: Optional[str] = None # If not set, defaults to the `spiders` directory under the current running script
    SPIDER_INTERCEPTORS_PATH: Optional[Union[ComponentInfo, Dict[str, int], List[str], str, None]] = ComponentInfo()
    DOWNLOAD_INTERCEPTORS_PATH: Optional[Union[ComponentInfo, Dict[str, int], List[str], str, None]] = ComponentInfo()
    ITEM_PIPELINES_PATH: Optional[Union[ComponentInfo, Dict[str, int], List[str], str, None]] = ComponentInfo()
    EXTENSIONS_PATH: Optional[Union[ComponentInfo, Dict[str, int], List[str], str, None]] = ComponentInfo()

    SCHEDULER: Optional[str] = None
    DUPEFILTER: Optional[str] = None
    BLOOM_INFO: Optional[BloomInfo] = BloomInfo()
    SCHEDULER_PERSIST: Optional[bool] = False
    DEDUP_TTL: Optional[int] = 0
    INCLUDE_HEADERS: Optional[List] = Field(default_factory=list) # Keys in headers to include during deduplication
    FILTER_KEY: Optional[str] = "cffiFilter"
    DONT_FILTER: Optional[bool] = False
    _new_seen: str = PrivateAttr()
    _sent_seeen: str = PrivateAttr()

    WS_END_TAG: Optional[str] = "websocket end" # You can customize the TAG to avoid conflicts with the response content
    # RET_COOKIES: Optional[Union[str, Literal[False]]] = "ret_cookies"  # False to disable cookie return; a string to specify the key used for returned cookies
    
    JS_PATH: Optional[Union[str, bool]] = None # Absolute/relative path to JS files or default to ./js_path under the running script directory
    
    LOG_INFO: Optional[LogInfo] = LogInfo()

    REDIS_INFO: Optional[RedisInfo] = RedisInfo()
    MYSQL_INFO: Optional[MysqlInfo] = MysqlInfo()
    MONBODB_INFO: Optional[MongodbInfo] = MongodbInfo()

    RABBITMQ_INFO: Optional[RabbitMQInfo] = RabbitMQInfo()
    KAFKA_INFO: Optional[KafkaInfo] = KafkaInfo()

    CPY_EXTENSIONS: Optional[CPYExtensionsConfig] = CPYExtensionsConfig()
    
    @property
    def _NEW_SEEN(self):
        return self._new_seen
    
    @property
    def _SENT_SEEN(self):
        return self._sent_seeen

    # Used to warn users about custom fields not recognized by the framework; these should be maintained by the user
    def __init__(self, **data: Any):
        super().__init__(**data)
        known_fields = set(self.model_fields.keys())
        extra_fields = set(data.keys()) - known_fields
        if extra_fields:
            import warnings
            warnings.warn(
                f"SettingsInfo received unrecognized fields: {extra_fields}. "
                "These fields will be ignored by the framework internals and should be managed by user code."
            )

    # Strict validation for component types
    @field_validator(
        "SPIDER_INTERCEPTORS_PATH",
        "DOWNLOAD_INTERCEPTORS_PATH",
        "ITEM_PIPELINES_PATH",
        "EXTENSIONS_PATH",
        mode="before"
    )
    @classmethod
    def validate_component(cls, v, info: ValidationInfo):
        if v is None or isinstance(v, ComponentInfo):
            return v
        return ComponentInfo.from_raw(v, info.field_name)
    
    @model_validator(mode='after')
    def check_after(self):
        if self.PROXY_URL:
            self.PROXIES = {"http": self.PROXY_URL, "https": self.PROXY_URL}

        self._new_seen  = f'{self.FILTER_KEY}_new_seen'
        self._sent_seeen = f'{self.FILTER_KEY}_sent_seen'
        return self
    
__all__ = [
    "LogInfo",
    "SettingsInfo"
]


import asyncio, time
from .redis import RedisScheduler
from ..downloader.internet import Request
from typing import TYPE_CHECKING, List
# from ...utils import run_with_timeout
from ...extensions import signals, SignalInfo
from ..sessions import SessionManager
if TYPE_CHECKING:
    from ...crawler import Crawler
    from ...settings import SettingsInfo
    from ...spiders import Spider
    from ...databases import RedisManager
    from ...extensions import SignalManager
    from ...mq.rabbitmq import RabbitMQManager

class RabbitMqScheduler(RedisScheduler):
    def __init__(
        self, 
        spiders_name: List=None,
        stop_event: asyncio.Event=None, 
        settings: "SettingsInfo"=None, 
        sessions: "SessionManager"=None, 
        sessions_lock: asyncio.Lock=None, 
        signalManager: "SignalManager"=None, 
        redisManager: "RedisManager"=None, 
        rabbitmqManager: "RabbitMQManager"=None,
        **kwargs
    ):
        super().__init__(
            spiders_name=spiders_name, 
            stop_event=stop_event, 
            settings=settings, 
            sessions=sessions, 
            sessions_lock=sessions_lock, 
            signalManager=signalManager, 
            redisManager=redisManager, 
            **kwargs
        )
        self.rabbitmqManager = rabbitmqManager

    @classmethod
    def from_crawler(cls, crawler: "Crawler", spiders_name: List):
        return cls(
            spiders_name=spiders_name, 
            stop_event=crawler.stop_event,
            settings=crawler.settings,
            sessions=crawler.sessions,
            sessions_lock=crawler.sessions_lock,
            signalManager=crawler.signalManager,
            redisManager=crawler.redisManager,
            rabbitmqManager=crawler.rabbitmqManager
        )
    
    async def put(self, request: "Request", spider: "Spider", **kwargs):
        is_seen = await self.dupefilter.request_seen(request=request, spider=spider)
        if is_seen:
            async with self.sessions_lock:
                self.sessions.release(session_id=request.session_id)
            self.signalManager.send(signal=signals.request_dropped, data=SignalInfo(signal_time=time.time(), request=request, reason=f"filter: {request.url}"))
            return False
        else:
            res = await self.rabbitmqManager.rpush(self.get_queue_key(spider=spider), request.to_bytes())
            if res:
                self.signalManager.send(signal=signals.request_scheduled, data=SignalInfo(signal_time=time.time(), request=request))
                return True
            else:
                async with self.sessions_lock:
                    self.sessions.release(session_id=request.session_id)
                self.signalManager.send(signal=signals.request_dropped, data=SignalInfo(signal_time=time.time(), request=request, reason=f"insert redis error: {request.url}"))
                return False

    async def get(self, spider: "Spider"=None, **kwargs):
        request_bytes = await self.rabbitmqManager.dequeue_request(queue_name=self.get_queue_key(spider=spider))
        if request_bytes is None:
            queue_size = await self.rabbitmqManager.llen(self.get_queue_key(spider=spider))
            return queue_size
        return Request.from_bytes(request_bytes)
    
    async def get_start_req(self, spider: "Spider", **kwargs):
        request_bytes = await self.rabbitmqManager.dequeue_request(queue_name=getattr(spider, "rabbitmq_queue", self.settings.QUEUE_NAME))
        if request_bytes is None:
            return None
        return request_bytes


from typing import Dict, Tuple
from parsel import Selector as ParselSelector
from typing import TYPE_CHECKING, Union, Dict, List
from functools import cached_property
from ...utils import extract_nested_objects, JSONScanner, ProtobufFactory
if TYPE_CHECKING:
    from .internet import HttpResponse

class Selector:
    def __init__(self, response: "HttpResponse"=None, type: str=None, **kwargs):
        if not response:
            raise ValueError("[Selector] Missing response")
        self.response = response
        self._text = getattr(response, "text", None)
        self._type = type or "html"

    @cached_property
    def _parsel_selector(self):
        if isinstance(self._text, str) and self._type in ("html", "xml"):
            return ParselSelector(text=self._text, type=self._type)
        return None
    
    def xpath(self, query):
        if self._parsel_selector:
            return self._parsel_selector.xpath(query)
        return []

    def css(self, query):
        if self._parsel_selector:
            return self._parsel_selector.css(query)
        return []

    def re(self, pattern):
        if self._parsel_selector:
            return self._parsel_selector.re(pattern)
        return []

    def json(self):
        return getattr(self.response, "json", lambda: None)()

    def extract_json(self, key: str="", re_rule: str="") -> Union[List[Union[Dict, str]], Dict, str]:
        if isinstance(self._text, str):
            return extract_nested_objects(text=self._text, key=key, re_rule=re_rule)
        return []

    def extract_json_strong(self, key=None, strict_level=2, re_rule="") -> Union[List[Union[Dict, str]], Dict, str]:
        if isinstance(self._text, str):
            return JSONScanner(strict_level=strict_level).scan_text(text=self._text, key=key, re_rule=re_rule)
        return []

    def protobuf_decode(self) -> Tuple[Dict, Dict]:
        content = getattr(self.response, "content", None)
        if content and isinstance(content, (bytes, bytearray)):
            return ProtobufFactory.protobuf_decode(self.response.content)
        return {}, {}
    
    def grpc_decode(self) -> Union[Tuple[Dict, Dict], List[Tuple[Dict, Dict]]]:
        content = getattr(self.response, "content", None)
        if content and isinstance(content, (bytes, bytearray)):
            return ProtobufFactory.grpc_decode(self.response.content)
        return {}, {}


import json, asyncio, sys
from .core.api import *
from .interceptors import ChainManager, InterruptibleChainManager
# from .interceptors import DownloadInterceptor
from .interceptors.api import UpdateRequestSpiderInterceptor, RobotSpiderInterceptor
from .pipelines import Pipeline
from .extensions import SignalManager
from .utils import load_object, get_class_name, get_all_spiders_cls, get_all_spiders_name, RobotsManager, get_run_py_dir, async_context_factory, KafkaLoggingHandler
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from .settings import SettingsInfo
    from logging import Logger

class Crawler:
    def __init__(self):
        self.run_py_dir = get_run_py_dir()
        self.stop_event = None
        self.global_lock = None

        self.settings: "SettingsInfo" = None
        self.scheduler = None
        self.taskManager: TaskManager = None
        self.downloader = None
        self.spiderInterceptor_chain = None
        self.downloadInterceptor_chain = None
        self.pipelines_chain = None
        self.sessions = None
        self.sessions_lock = None

        self.redisManager = None
        self.mysqlManager = None
        self.mongodbManager = None
        self.rabbitmqManager = None
        self.kafkaManager = None

        self.logger: "Logger" = None
        self.signalManager = None
        self.robot = None
        self.extensions_list = []

        self.spiders = None
        self.engines = None

    def init_output(self, class_list):
        return [get_class_name(it) for it in class_list] if isinstance(class_list, list) else [get_class_name(class_list)]

    async def do_initialization(self, settings: "SettingsInfo", start_type=0):
        self.stop_event = asyncio.Event()

        self.settings: "SettingsInfo" = settings
        from .cpy import CExtensionLoader
        from .models.api import CPYExtension

        framework_cpy = [
            CPYExtension(module_name="bloom")
        ]
        framework_cpy.extend(self.settings.CPY_EXTENSIONS.RESOURCES) # User first principle, same name can cover framework modules
        self.settings.CPY_EXTENSIONS.RESOURCES = framework_cpy
        CExtensionLoader(resource_dir=self.settings.CPY_EXTENSIONS.DIR).load_all(configs=self.settings.CPY_EXTENSIONS.RESOURCES)

        self.global_lock = async_context_factory(
            max_tasks=self.settings.MAX_GLOBAL_CONCURRENT_TASKS,
            semaphore_cls=asyncio.BoundedSemaphore
        )

        # if not logger: # To ensure log stability, it is no longer enabled
        from .utils import init_logger
        logger = init_logger(log_info=self.settings.LOG_INFO, logger_name=__name__)
        self.logger = logger
        # kafka
        if self.settings.KAFKA_INFO.resolved_url:
            from .mq.kafka import KafkaManager
            self.kafkaManager = KafkaManager.from_crawler(self)
            kafka_handler = KafkaLoggingHandler(kafka=self.kafkaManager, stop_event=self.stop_event).create_fmt(self.settings)
            self.logger.addHandler(kafka_handler)

        self.sessions_lock = asyncio.Lock()
        self.sessions = SessionManager.from_crawler(self)
        self.signalManager = SignalManager.from_crawler(self)
        self.robot = RobotsManager.from_crawler(self)
        
        # redis
        if self.settings.REDIS_INFO.resolved_url:
            from .databases import RedisManager
            self.redisManager = RedisManager.from_crawler(self)

        # mysql
        if self.settings.MYSQL_INFO.resolved_url:
            from .databases.mysql import SQLAlchemyMySQLManager
            self.mysqlManager = SQLAlchemyMySQLManager.from_crawler(self)

        # mongodb
        if self.settings.MONBODB_INFO.resolved_url:
            from .databases.mongodb import MongoDBManager
            self.mongodbManager = MongoDBManager.from_crawler(self)

        # rabbitmq
        if self.settings.RABBITMQ_INFO.resolved_url:
            from .mq.rabbitmq import RabbitMQManager
            self.rabbitmqManager = RabbitMQManager.from_crawler(self)
            if not self.settings.SCHEDULER:
                self.settings.SCHEDULER = "scrapy_cffi.scheduler.RabbitMqScheduler"

        self.settings.SPIDER_INTERCEPTORS_PATH.value.extend([RobotSpiderInterceptor, UpdateRequestSpiderInterceptor])
        self.spiderInterceptor_chain = InterruptibleChainManager.from_crawler(self, class_list=self.settings.SPIDER_INTERCEPTORS_PATH.value)

        # self.settings.DOWNLOAD_INTERCEPTORS_PATH.value.insert(0, DownloadInterceptor)
        self.downloadInterceptor_chain = InterruptibleChainManager.from_crawler(self, class_list=self.settings.DOWNLOAD_INTERCEPTORS_PATH.value)

        self.settings.ITEM_PIPELINES_PATH.value.insert(0, Pipeline)
        self.pipelines_chain = ChainManager.from_crawler(self, class_list=self.settings.ITEM_PIPELINES_PATH.value)

        from .hooks import signals_hooks
        for ext_cls in self.settings.EXTENSIONS_PATH.value:
            self.extensions_list.append(ext_cls.from_crawler(
                hooks=signals_hooks(self), 
                redisManager=self.redisManager,
                mysqlManager=self.mysqlManager,
                mongodbManager=self.mongodbManager,
                rabbitmqManager=self.rabbitmqManager,
                kafkaManager=self.kafkaManager,
        ))

        self.downloader = Downloader.from_crawler(self)

        # spider start type
        if not self.settings.SPIDERS_PATH:
            self.settings.SPIDERS_PATH = str(self.run_py_dir / "spiders")
            self.logger.warning(f"not provided self.settings.SPIDERS_PATH，guessed to load -> {self.settings.SPIDERS_PATH}")
            start_type = 0
        if start_type:
            self.spiders = [load_object(path=self.settings.SPIDERS_PATH)]
            spiders_name = [spider.name for spider in self.spiders]
        else:
            self.spiders = get_all_spiders_cls(spiders_dir=self.settings.SPIDERS_PATH)
            spiders_name = get_all_spiders_name(logger=self.logger, spiders_cls_list=self.spiders)

        scheduler_path = self.settings.SCHEDULER
        if scheduler_path:
            scheduler_cls = load_object(path=scheduler_path)
        else:
            from .core.scheduler import Scheduler
            scheduler_cls = Scheduler
        self.scheduler = scheduler_cls.from_crawler(self, spiders_name)

        for spider_cls in self.spiders:
            has_redis_key = getattr(spider_cls, "redis_key", None)
            if has_redis_key:
                self.taskManager = TaskManager.from_crawler(self, is_distributed=True) # Shared by all spider engines; if any exist, start one to handle blocking
                break
        else:
            self.taskManager = TaskManager.from_crawler(self, is_distributed=False)

        robot_task = None
        if self.settings.ROBOTSTXT_OBEY:
            robot_urls = set()
            for spider_cls in self.spiders:
                scheme = getattr(spider_cls, "robot_scheme", "https").lower()
                for domain in getattr(spider_cls, "allowed_domains", []):
                    robot_urls.add(f"{scheme}://{domain}/robots.txt")
            now_loop = asyncio.get_running_loop()
            robot_task = now_loop.create_task(self.robot.load_rules_for_hosts(robot_urls))

        self.spiders = [spider.from_crawler(self) for spider in self.spiders]
        self.engines = [Engine.from_crawler(crawler=self, spider=spider) for spider in self.spiders]

        core_data = []
        for spider, engine in zip(self.spiders, self.engines):
            core_data.append({"spider": self.init_output(spider)[0], "engine": self.init_output(engine)[0]})
    
        init_data = {
            "taskManager": self.init_output(self.taskManager)[0],
            "sessions": self.init_output(self.sessions)[0],
            "scheduler": self.init_output(self.scheduler)[0],
            "downloader": self.init_output(self.downloader)[0],
            "spiderInterceptor_chain": self.init_output(self.settings.SPIDER_INTERCEPTORS_PATH.value),
            "downloadInterceptor_chain": self.init_output(self.settings.DOWNLOAD_INTERCEPTORS_PATH.value),
            "pipelines_chain": self.init_output(self.settings.ITEM_PIPELINES_PATH.value),
            "extensions": self.init_output(self.extensions_list),
            "core": core_data
        }
        init_text = json.dumps(init_data, indent=4, ensure_ascii=False)
        self.logger.debug(init_text)
        return robot_task
    
    async def start_engines(self, robot_task, *args, **kwargs):
        self.signalManager.start()
        self.sessions.start()
        if robot_task:
            await robot_task
        await asyncio.gather(*[engine.start(*args, **kwargs) for engine in self.engines])
        self.stop_event.set()
        await self.sessions.close_all()
        await self.signalManager.stop()

    async def shutdown(self):
        self.stop_event.set()
        self.taskManager.active_tasks = 0
        self.taskManager.tasks_done_event.set()
        self.taskManager.error_event.set()

        # await asyncio.sleep(1)
        for engine in self.engines:
            await engine.taskManager.cancel_all()
        await self.sessions.close_all()
        await self.signalManager.stop()

        if not self.settings.SCHEDULER_PERSIST and self.redisManager:
            for spider in self.spiders:
                if getattr(spider, "redis_key", None):
                    await self.redisManager.delete(spider.redis_key)
            await self.redisManager.delete(self.settings.QUEUE_NAME)
            await self.redisManager.delete(self.settings._NEW_SEEN)
            await self.redisManager.delete(self.settings._SENT_SEEN)
        
        if self.rabbitmqManager:
            await self.rabbitmqManager.close()

        if self.kafkaManager:
            await self.kafkaManager.close()

def cleanup_loop(loop: asyncio.AbstractEventLoop):
    pending = asyncio.all_tasks(loop=loop)
    for task in pending:
        task.cancel()
        try:
            loop.run_until_complete(task)
        except asyncio.CancelledError:
            pass
    loop.close()

# One spider corresponds to one engine
async def run_base(start_type, settings: "SettingsInfo", new_loop=False, *args, **kwargs):
    if new_loop: # Suitable for running in an independent thread or synchronous start, use with caution to avoid cross event loop operations
        now_loop = asyncio.new_event_loop()
        asyncio.set_event_loop(now_loop)
    else: # Suitable for calling within an existing asynchronous environment (default)
        now_loop = asyncio.get_running_loop()
    crawler = Crawler()
    robot_task = await crawler.do_initialization(settings=settings, start_type=start_type)
    engine_task = now_loop.create_task(crawler.start_engines(robot_task=robot_task, *args, **kwargs))
    return crawler, engine_task

def run_sync_base(start_type, settings: "SettingsInfo", new_loop=True, *args, **kwargs):
    if new_loop:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
    else:
        loop = asyncio.get_running_loop()
    crawler: Crawler = None

    async def main():
        nonlocal crawler
        crawler = Crawler()
        robot_task = await crawler.do_initialization(settings=settings, start_type=start_type)
        await crawler.start_engines(robot_task, *args, **kwargs)
        crawler.stop_event.set()
        await crawler.shutdown()

    try:
        loop.run_until_complete(main())
    except KeyboardInterrupt:
        print("KeyboardInterrupt detected, shutting down...")
        if crawler:
            loop.run_until_complete(crawler.shutdown())
    finally:
        cleanup_loop(loop=loop)

# Run a single spider, where one set of components corresponds to one spider
async def run_spider(settings: "SettingsInfo", new_loop=False, *args, **kwargs):
    return await run_base(start_type=1, settings=settings, new_loop=new_loop, *args, **kwargs)

# Run all spiders, where one set of components corresponds to multiple spiders
async def run_all_spiders(settings: "SettingsInfo", new_loop=False, *args, **kwargs):
    return await run_base(start_type=0, settings=settings, new_loop=new_loop, *args, **kwargs)

def run_spider_sync(settings: "SettingsInfo", new_loop=True, *args, **kwargs):
    return run_sync_base(start_type=1, settings=settings, new_loop=new_loop, *args, **kwargs)

def run_all_spiders_sync(settings: "SettingsInfo", new_loop=True, *args, **kwargs):
    return run_sync_base(start_type=0, settings=settings, new_loop=new_loop, *args, **kwargs)

__all__ = [
    "Crawler",
    "cleanup_loop",
    "run_spider",
    "run_all_spiders",
    "run_spider_sync",
    "run_all_spiders_sync",
]


# exception.py
from typing import Any, Optional, List

class BlackboxProtobufException(Exception):
    """Base class for excepions raised by Blackbox Protobuf"""

    def __init__(self, message: str, path: Optional[List[str]]=None, *args) -> None:
        self.path = path
        super().__init__(message, *args)

    def set_path(self, path: List[str]) -> None:
        if self.path is None:
            self.path = path


class TypedefException(BlackboxProtobufException):
    """Thrown when an error is identified in the type definition, such as
    conflicting or inconsistent values."""

    def __str__(self):
        message = super().__str__()
        if self.path is not None:
            message = (
                "Encountered error within typedef for field %s: "
                % "->".join(map(str, self.path))
            ) + message
        else:
            message = ("Encountered error within typedef: ") + message
        return message


class EncoderException(BlackboxProtobufException, ValueError):
    """Thrown when there is an error encoding a dictionary to a type definition"""

    def __str__(self):
        message = super().__str__()
        if self.path is not None:
            message = (
                "Encountered error encoding field %s: " % "->".join(map(str, self.path))
            ) + message
        else:
            message = ("Encountered error encoding message: ") + message
        return message


class DecoderException(BlackboxProtobufException, ValueError):
    """Thrown when there is an error decoding a bytestring to a dictionary"""

    def __str__(self):
        message = super().__str__()
        if self.path is not None:
            message = (
                "Encountered error decoding field %s: " % "->".join(map(str, self.path))
            ) + message
        else:
            message = ("Encountered error decoding message: ") + message
        return message


class ProtofileException(BlackboxProtobufException):
    def __init__(self, message: str, path: Optional[List[str]]=None, filename: Optional[str]=None, *args) -> None:
        self.filename = filename
        super().__init__(message, path, *args)

    def __str__(self):
        message = super().__str__()
        if self.path is not None:
            message = (
                "Encountered error within protofile %s for field %s: "
                % (self.filename, "->".join(map(str, self.path)))
            ) + message
        else:
            message = (
                "Encountered error within protofile %s: " % self.filename
            ) + message

        return message



from .request import Request, HttpRequest, MediaRequest, WebSocketRequest
from .response import Response, HttpResponse, WebSocketResponse

__all__ = [
    "Request",
    "HttpRequest",
    "MediaRequest",
    "WebSocketRequest",
    "Response",
    "HttpResponse",
    "WebSocketResponse"
]


import random
from scrapy_cffi.spiders import Spider
from scrapy_cffi.internet import *
from copy import deepcopy

class StudentSpider(Spider):
    name = "student"
    robot_scheme = "http"
    allowed_domains = ["127.0.0.1:8002", "127.0.0.1:8765"]

    school_id = int(random.random() * 100)
    count = 0
    start_urls = [f"http://127.0.0.1:8002/school/{school_id}"]

    async def parse(self, response: HttpResponse):
        response_json = response.json()
        if response_json.get("data", []):
            class_list = response_json["data"]
            for s_class in class_list:
                class_id = s_class["class_id"]
                student_ids = s_class["student_ids"]
                for tid in student_ids:
                    # The backend server is a minimal mock service with poorly designed APIs.
                    # To avoid accidental ID deduplication (e.g., overlapping student IDs across classes),
                    # we construct a unique absolute ID by combining class_id and student_id.
                    # This ensures IDs are distinct even if the original IDs are reused.
                    abs_tid = self.count + class_id * 100 + tid
                    yield HttpRequest(
                        session_id=response.session_id,
                        url=f"http://127.0.0.1:8002/student/{abs_tid}",
                        method="GET",
                        callback=self.parse_student,
                        errback=self.errRet,
                        meta=deepcopy({"class_id": class_id, "student_id": abs_tid})
                    )

            has_next = response_json.get("has_next")
            if has_next:
                self.count += 1000000
                yield HttpRequest(
                    session_id=response.session_id,
                    url=self.start_urls[0],
                    method="GET",
                    params={"cursor": response_json["cursor"]},
                    callback=self.parse,
                    errback=self.errRet
                )
        else:
            print(response.json())
    
    async def parse_student(self, response: HttpResponse):
        meta = response.meta
        response_json = response.json()
        meta.update({"student_info": response_json})

        img_url = f'http://127.0.0.1:8002{response_json["img"]}'
        yield HttpRequest(
            session_id=response.session_id,
            url=img_url,
            callback=self.collect_data,
            errback=self.errRet,
            meta=deepcopy(meta)
        )

    async def collect_data(self, response: HttpResponse):
        meta = response.meta
        response_content = response.content
        meta["student_info"]["t_pic_data"] = response_content
        yield meta